/**********************************************
* Copyright (c) 2021-2022 Crypto Quantique Ltd 
***********************************************/

#include <stdio.h>
#include <stddef.h>

#include "../Inc/qlLibrary.h"
#include "cloud_wifi_task.h"
#include "atcacert/atcacert_pem.h"
#include "cloud_wifi_ecc_process.h"
#include "tng/tng_atca.h"
#include "hex_dump.h"


#define CSRDEBUG			0U	// 0 = no CSR debug 

#define PUBKEY_HEADER_LEN	27U
#define PUBKEY_LEN			64U
#define PUBKEY_DER_LEN		PUBKEY_HEADER_LEN+PUBKEY_LEN

#define SHA_256_LENGTH      32U

// urlEncode defs
static const unsigned char pucHex[16] = "0123456789abcdef";
#define URL_ENCODE_INVALID_PARAM        (60000)
#define URL_ENCODE_BUFFER_OVERFLOW      (60010)
#define INLINE inline


// Local function definitions
int gen_CertificateRequest(char* deviceID, int slot, uint8_t *publicKeyDer,int derLen,char *CSRcertPem) ;
int buildCSR(char *CommonName, uint8_t *PubKeyDer, uint8_t *csrDer);
int makeASN1Sig(int offset,uint8_t *stored_Response, uint8_t *csrDer);
int addSignature(int offset, uint8_t *stored_Response, uint8_t *csrDer);

static INLINE int add_char(char **ppcBuf, const char * const pcEnd, unsigned char uc);


int createESKCSR(char *pcCSRout, char* deviceID,uint8_t *publicKeyDer,size_t *pxOutWritten) {
	int rc = 0;
    char lc_deviceID[65];
    
    // make local copy 
    strcpy(lc_deviceID,deviceID);
    
	rc = gen_CertificateRequest(lc_deviceID,SLOT_0, publicKeyDer,
			PUBKEY_DER_LEN, pcCSRout);
	*pxOutWritten = (size_t) strlen(pcCSRout);

	if (CSRDEBUG)
		APP_DebugPrintf("%s\n\n",pcCSRout);
	return rc;
}

int createDSKCSR(char *pcCSRout, char* deviceID,uint8_t *publicKeyDer,size_t *pxOutWritten) {
	int rc = 0;
    char lc_deviceID[65];
    
    // make local copy 
    strcpy(lc_deviceID,deviceID);
    
	rc = gen_CertificateRequest(lc_deviceID,SLOT_2, publicKeyDer,
			PUBKEY_DER_LEN, pcCSRout);
	*pxOutWritten = (size_t) strlen(pcCSRout);

	if (CSRDEBUG)
		APP_DebugPrintf("\n\n%s\n\n",pcCSRout);
	return rc;
}


int gen_CertificateRequest(char* deviceID, int slot, uint8_t *publicKeyDer, 
        int derLen,char *CSRcertPem) 
{
	int rc = 0;
	ATCA_STATUS status = ATCA_SUCCESS;
    uint8_t csrDer[300] = { 0 };
	uint8_t data_Hash[SHA_256_LENGTH] = { 0 };
	uint8_t sig[64] = { 0 };
    size_t outlen =500;

	if (deviceID == NULL)
		APP_DebugPrintf("Warning DeviveID not set\n");

	if (CSRDEBUG)
		APP_DebugPrintf("\r\nDBGS qlDeviceID = \"%s\"\n",deviceID);

	// Generate first Part of CSR
	derLen = buildCSR(deviceID, publicKeyDer, csrDer);

	//  data_hash of Cert so far excluding first 4 bytes of CSR header
    if (_gDevice==NULL)
    {
    APP_DebugPrintf("HW SHA256 _gDevice Not defined.\r\n");  
    return -1;
    }

status = calib_hw_sha2_256(_gDevice, csrDer + 4, derLen - 4, data_Hash);

if (status != ATCA_SUCCESS)
    APP_DebugPrintf("hw SHA256 failed.\r\n");
else
    {
    // convert the hash binary to hex chars
    for (int i = 0; i < sizeof(data_Hash); i++)
    {
        sprintf(&deviceID[2 * i], "%02x", data_Hash[i]);
    }
    deviceID[64] = '\0';
    }

	if (CSRDEBUG) {
		APP_DebugPrintf(("Data to be signed \n"));
		print_hex_dump(data_Hash, sizeof(data_Hash), FALSE,FALSE,16);
	}
        
    status = atcab_sign(slot, data_Hash, sig);

    if (status != ATCA_SUCCESS){
        APP_DebugPrintf("Error : %d Failed to get CSR signature.\n",status);
    }
    
	if (CSRDEBUG) {
		APP_DebugPrintf(("Signature : \n"));
        print_hex_dump(sig, sizeof(sig), FALSE,FALSE,16); 
	}

	// sign CSR with Response
	derLen = addSignature(derLen, sig, csrDer);

	if (CSRDEBUG > 1) {
		APP_DebugPrintf(("CSR DER :\n"));
		print_hex_dump(csrDer, derLen, FALSE,FALSE,16);
	}

    if (rc==0)
    {
        rc = atcacert_encode_pem_csr(csrDer, derLen, CSRcertPem, &outlen);
    }
	return rc;
}


int buildCSR(char *CommonName, uint8_t *PubKeyDer, uint8_t *csrDer) {
	uint8_t csrHeader[] = { 0x30, 0x82, 0x01, 0x05, 0x30, 0x81, 0xAD, 0x02,
			0x01, 0x00, 0x30, 0x4B, 0x31, 0x49, 0x30, 0x47, 0x06, 0x03, 0x55,
			0x04, 0x03, 0x13, 0x40 };

	int len = 0;
	int offset = 0;

	len = sizeof(csrHeader);
	memcpy(csrDer, &csrHeader, len);
	offset += len;

	len = strlen(CommonName);
	memcpy(csrDer + offset, CommonName, len);
	offset += len;

	uint8_t part2[] = { 0x30, 0x59, 0x30, 0x13, 0x06, 0x07, 0x2A, 0x86, 0x48,
			0xCE, 0x3D, 0x02, 0x01, 0x06, 0x08, 0x2A, 0x86, 0x48, 0xCE, 0x3D,
			0x03, 0x01, 0x07, 0x03, 0x42, 0x0, 0x04 };

	len = sizeof(part2);
	memcpy(csrDer + offset, &part2, len);
	offset += len;

	len = PUBKEY_LEN;
	memcpy(csrDer + offset, PubKeyDer + PUBKEY_HEADER_LEN, len);
	offset += len;

	csrDer[offset++] = 0xA0;
	csrDer[offset++] = 0x00;

	return offset;
}

void incrementLen(uint8_t *csrDer) { csrDer[3]++; csrDer[193]++; csrDer[196]++; }
void decrementLen(uint8_t *csrDer) { csrDer[3]--; csrDer[193]--; csrDer[196]--; }

int makeASN1Sig(int offset,uint8_t *stored_Response, uint8_t *csrDer)
{
	int i=0;

	// if first byte Zero
	// decrement lengths and
	// use sig[1] thou sig[32]
	if (stored_Response[i] == 0x0)
			{
			decrementLen(csrDer);
			csrDer[offset++] = 0x1F;
			i++;

			// if first byte >127 extend
			if (stored_Response[i] & 0x80)
				{
                incrementLen(csrDer);
				offset--; // go back to modify length field
				csrDer[offset++] = 0x20;
				csrDer[offset++] = 0x00;
				}
			}
	else

	// if first byte >127 extend
	if (stored_Response[i] & 0x80)
			{
			incrementLen(csrDer);
			csrDer[offset++] = 0x21;
			csrDer[offset++] = 0x00;
			}
	else
			csrDer[offset++] = 0x20;


	for (; i<32; i++){
		csrDer[offset++] = stored_Response[i];
	}

return offset;
}


int addSignature(int offset, uint8_t *stored_Response, uint8_t *csrDer) {
	
	uint8_t part3[] = { 0x30, 0x0A, 0x06, 0x08, 0x2A, 0x86, 0x48, 0xCE, 0x3D,
			0x04, 0x03, 0x02, 0x03, 0x47, 0x00, 0x30, 0x44 };

	int len = sizeof(part3);
	memcpy(csrDer + offset, &part3, len);
	offset += len;

	//
	// adding sig 'r'
	csrDer[offset++] = 0x02;

	offset = makeASN1Sig(offset,(uint8_t *)stored_Response,csrDer);

	//
	// adding sig 's'
	csrDer[offset++] = 0x02;

	offset = makeASN1Sig(offset,(uint8_t *)stored_Response+32,csrDer);

	return offset;
}




/** Add a character to the supplied buffer, advancing its pointer
 *
 * Safe to call multiple times: this function guards against overrun
 *
 * @note After a successful return, *ppcBuf is incremented
 *
 * @note @p uc is supplied unsigned, despite buffers being chars, for
 *            convenience in @p url_encode()
 *
 * @param[out]	 ppcBuf	 Ptr to the buffer ptr to which to add the character
 * @param[in]    pcEnd   Limit of that buffer
 * @param[in]    uc      Character to add to the buffer
 * @return 0 on success, non-0 if @p *ppcBuf was already >= @p pcEnd
 */
static INLINE int add_char(char **ppcBuf, const char * const pcEnd, unsigned char uc)
{
  int ret = (*ppcBuf != pcEnd) ? 0 : URL_ENCODE_BUFFER_OVERFLOW;

  if (0 == ret)
  {
	*((*ppcBuf)++) = (char)uc;
  }

  return ret;
}


int urlEncode(const char *pcIn, size_t xInLen,
			  char *pcOut, size_t xOutLen, size_t *pxOutWritten)
{
  if (   NULL == pcIn
	  || 0 == xInLen
	  || NULL == pcOut
	  || 0 == xOutLen
	  || NULL == pxOutWritten)
  {
	return URL_ENCODE_INVALID_PARAM;
  }

  const char *pcInEnd = pcIn + xInLen;
  const char *pcOutStart = pcOut;
  const char *pcOutEnd = pcOut + xOutLen;

  int ret = 0;

  while (0 == ret && pcIn < pcInEnd)
  {
  	/* Cast to unsigned char for arithmetic & shifts */
	unsigned char c = (unsigned char)(*pcIn++);

	if (!isalnum(c))
	{
      switch (c)
      {
		/* There is inconsistency between RFC 3986 and HTML 5 .4.10.22.5
		 * over the correct handling of, ' ', '*' and '~'. We follow the
		 * HTML 5 specification here.
		 */
		case '-':
		case '.':
		case '_':
		case '*':
			/* Leave unchanged */
			break;
		case ' ':
			c = '+';
			break;
		default:
			/* Hexify the character. Safe even for buffer-full. The
			 * final add_char(), after the if() statement, will catch
			 * buffer-full and terminate the loop.
			 */
			(void)add_char(&pcOut, pcOutEnd, '%');
			(void)add_char(&pcOut, pcOutEnd, pucHex[((c >> 4) & 0xf)]);

			/* And reuse the final add_char() */
			c = pucHex[c & 0xf];
			break;
      }
	}

	ret = add_char(&pcOut, pcOutEnd, c);
  }
  *pcOut='\0';
  *pxOutWritten = (size_t)(pcOut - pcOutStart);

  return ret;
}


