/**********************************************
* Copyright (c) 2021-2022 Crypto Quantique Ltd
***********************************************/


#include <stdio.h>


#include "../Inc/qlLibrary.h"
#include "cloud_wifi_task.h"
#include "atcacert/atcacert_pem.h"
#include "cloud_wifi_ecc_process.h"
#include "tng/tng_atca.h"


int parseEnrolResponse(char *pcBody, struct ProvisionedDetails *pPdOut);
int storeIoTdetail(char *details, int len);
int8_t separateSigDev(char* listCerts,char* sCertPEM,char* dCertPEM);
int8_t transfer_certificates(uint8_t* deviceCert,size_t deviceCertLen,
                                uint8_t* signerCert,size_t signerCertLen);
int8_t set_ssl_callback(void);

int provision(char *pcEnrolResponse,char *pcUrl,uint16_t* sPortNum)
{
  int8_t rc;
  struct ProvisionedDetails pdProv;

  char    dCertPEM[1024];
  char    sCertPEM[1024];

  uint8_t deviceCert[1024];
  size_t  deviceCertLen = 1024;

  uint8_t signerCert[1024];
  size_t  signerCertLen = 1024;


  pcEnrolResponse = strstr(pcEnrolResponse,PEM_CERT_BEGIN);

  rc = parseEnrolResponse (pcEnrolResponse, &pdProv);
  if (rc != 0)
  {
    APP_DebugPrintf("**** provision: parseEnrolResponse failed, rc=%i\r\n", rc);
    return rc;
  }

   strcpy(pcUrl,pdProv.iotHubURL);
   *sPortNum = pdProv.iotHubPort;

   rc = separateSigDev(pdProv.dcrt,sCertPEM,dCertPEM);

   if (rc != 0)
    {
     APP_DebugPrintf("**** failed, dcrt response rc=%i\r\n", rc);
     return rc;
    }

   rc = atcacert_decode_pem(dCertPEM,strlen(dCertPEM),
           deviceCert,&deviceCertLen,
           PEM_CERT_BEGIN,
           PEM_CERT_END);

   if (rc != 0)
   {
    APP_DebugPrintf("**** provision: pem decode failed, rc=%i\r\n", rc);
    return rc;
   }


   rc = atcacert_decode_pem(sCertPEM,strlen(sCertPEM),
           signerCert,&signerCertLen,
           PEM_CERT_BEGIN,
           PEM_CERT_END);

   if (rc != 0)
   {
    APP_DebugPrintf("**** provision: pem decode failed, rc=%i\r\n", rc);
    return rc;
   }

   rc = transfer_certificates(deviceCert,deviceCertLen,signerCert,signerCertLen);

   set_global_slot_index(SLOT_2);
   set_ssl_callback();

   if (rc != M2M_SUCCESS)
   {
    APP_DebugPrintf("**** provision: winc Certs transfer failed , rc=%i\r\n", rc);
    return rc;
   }

   // save IoT Hub URL and Port
   char details[100];
   sprintf(details,"%s:%d\n",pdProv.iotHubURL,pdProv.iotHubPort);
   storeIoTdetail(details,strlen(details));

  return rc;
}

int8_t separateSigDev(char* listCerts,char* sCertPEM,char* dCertPEM)
{
 char *ptr = strstr(listCerts,PEM_CERT_END)+strlen(PEM_CERT_END)+1;

    if (ptr<0)
       return -1;

   strcpy(sCertPEM,ptr);
   if (strlen(sCertPEM)<100)
       return -2;

   *ptr = '\0';
   strcpy(dCertPEM,listCerts);

   if (strlen(dCertPEM)<100)
       return -3;

   return 0;
}



int getDeviceId(char* deviceID)
{
ATCA_STATUS status = ATCA_SUCCESS;
uint8_t public_key_DER[91];
uint8_t pubKeyHash[32];

status = getPublicKey(public_key_DER,SLOT_0);

if (_gDevice==NULL)
    {
    APP_DebugPrintf("HW SHA256 _gDevice Not defined.\r\n");
    return -1;
    }

status = calib_hw_sha2_256(_gDevice, public_key_DER, sizeof(public_key_DER), pubKeyHash);

if (status!=ATCA_SUCCESS)
    APP_DebugPrintf("hw SHA256 failed.\r\n");
else
    {
    // convert the hash binary to hex chars
    for (int i = 0; i < sizeof(pubKeyHash); i++)
    {
        sprintf(&deviceID[2 * i], "%02x", pubKeyHash[i]);
    }
    deviceID[64] = '\0';
    }

return status;
}

int getPublicKey(uint8_t *public_key_DER, uint32_t slot)
{
    ATCA_STATUS status = ATCA_SUCCESS;
    uint8_t public_key[64];

    uint8_t PublicKeyHeader[] = {
            0x30, 0x59, 0x30, 0x13, 0x06, 0x07, 0x2a, 0x86,
            0x48, 0xce, 0x3d, 0x02, 0x01, 0x06, 0x08, 0x2a,
            0x86, 0x48, 0xce, 0x3d, 0x03, 0x01, 0x07, 0x03,
            0x42, 0x00, 0x04
            };

    // set up header
    memcpy(public_key_DER,PublicKeyHeader,sizeof(PublicKeyHeader));

    if (slot==SLOT_0)
        status = tng_get_device_pubkey(public_key);
    else
        status = atcab_genkey(slot, public_key);

    if (status!=ATCA_SUCCESS)
        {
        APP_DebugPrintf("get Public key failed.\r\n");
        return status;
        }

    memcpy(public_key_DER+sizeof(PublicKeyHeader),public_key,sizeof(public_key));

    return status;
}


// Define a constant array in Flash.
// It must be aligned to row boundary and size has to be in multiple of rows
const uint8_t nvm_user_start_address[NVMCTRL_FLASH_ROWSIZE] __attribute__((aligned(NVMCTRL_FLASH_ROWSIZE),keep,externally_visible,space(prog)))= {0};




int storeIoTdetail(char *details, int len) {
    uint8_t data [NVMCTRL_FLASH_PAGESIZE] = {0};

    if (len>NVMCTRL_FLASH_PAGESIZE){
        APP_DebugPrintf("Error : NVM write.\r\n");
        return -1;
        }

    memcpy(data,(uint8_t *)details,len);

    /* Erase the row */
    NVMCTRL_RowErase((uint32_t)nvm_user_start_address);
    while(NVMCTRL_IsBusy());

    /* Program 64 byte page */
    NVMCTRL_PageWrite((uint32_t *)data, (uint32_t)nvm_user_start_address);
    while(NVMCTRL_IsBusy());

     /* Verify the programmed content*/
    if (!memcmp(data, (void *)nvm_user_start_address, NVMCTRL_FLASH_PAGESIZE))
    {
        APP_DebugPrintf("NVM Write Success.\r\n");
    }
    else
        APP_DebugPrintf("NVM Write Failed.\r\n");

    return 0;
}

int retrieveIoTdetail(char *pcUrl,uint16_t *sPortNum) {
    uint8_t data [NVMCTRL_FLASH_PAGESIZE] = {0};
    char details[256] = {0};

      memcpy(data, (void *)nvm_user_start_address, NVMCTRL_FLASH_PAGESIZE) ;
      strcpy(details, (char *)data);

    if(strlen(details)>0){
        *sPortNum = (uint16_t)parse(details,":","",1);

        char *ptr = strstr(details,":");
        *ptr = '\0';
        strcpy(pcUrl,details);
        }
    return strlen(details);
}

int removeIoTdetail() {

    /* Erase the row */
    NVMCTRL_RowErase((uint32_t)nvm_user_start_address);
    while(NVMCTRL_IsBusy());

    return 0;
}