/**********************************************
* Copyright (c) 2021-2022 Crypto Quantique Ltd 
***********************************************/

#include <app.h>

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "drv/driver/m2m_types.h"
#include "drv/driver/m2m_ssl.h"
#include "atcacert/atcacert_client.h"
#include "../drv/common/nm_common.h"


#define MAX_TLS_CERT_LENGTH         1024
#define SIGNER_CERT_MAX_LEN         600 //Set Maximum of existing certs
#define DEVICE_CERT_MAX_LEN         600
#define CERT_SN_MAX_LEN             32
#define TLS_SRV_ECDSA_CHAIN_FILE    "ECDSA.lst"
#define INIT_CERT_BUFFER_LEN        (MAX_TLS_CERT_LENGTH*sizeof(uint32_t) - TLS_FILE_NAME_MAX*2 - SIGNER_CERT_MAX_LEN - DEVICE_CERT_MAX_LEN)



static size_t winc_certs_get_total_files_size(const tstrTlsSrvSecHdr* header)
{
    uint8_t *pBuffer = (uint8_t*)header;
    uint16_t count = 0;

    while ((*pBuffer) == 0xFF)
    {

        if (count == INIT_CERT_BUFFER_LEN)
        {
            break;
        }
        count++;
        pBuffer++;
    }

    if (count == INIT_CERT_BUFFER_LEN)
    {
        return sizeof(*header); // Buffer is empty, no files

    }
    return header->u32NextWriteAddr;
}


static int8_t winc_certs_append_file_buf(uint32_t* buffer32, uint32_t buffer_size,
                                         const char* file_name, uint32_t file_size,
                                         const uint8_t* file_data)
{
    tstrTlsSrvSecHdr* header = (tstrTlsSrvSecHdr*)buffer32;
    tstrTlsSrvSecFileEntry* file_entry = NULL;
    uint16_t str_size = (uint8_t)strlen((char*)file_name) + 1;
    uint16_t count = 0;
    uint8_t *pBuffer = (uint8_t*)buffer32;

    while ((*pBuffer) == 0xFF)
    {

        if (count == INIT_CERT_BUFFER_LEN)
        {
            break;
        }
        count++;
        pBuffer++;
    }

    if (count == INIT_CERT_BUFFER_LEN)
    {
        // The WINC will need to add the reference start pattern to the header
        header->u32nEntries = 0;                    // No certs
        // The WINC will need to add the offset of the flash were the certificates are stored to this address
        header->u32NextWriteAddr = sizeof(*header); // Next cert will be written after the header
    }

    if (header->u32nEntries >= sizeof(header->astrEntries) / sizeof(header->astrEntries[0]))
    {
        return M2M_ERR_FAIL; // Already at max number of files

    }
    if ((header->u32NextWriteAddr + file_size) > buffer_size)
    {
        return M2M_ERR_FAIL; // Not enough space in buffer for new file

    }
    file_entry = &header->astrEntries[header->u32nEntries];
    header->u32nEntries++;

    if (str_size > sizeof(file_entry->acFileName))
    {
        return M2M_ERR_FAIL; // File name too long
    }
    memcpy((uint8_t*)file_entry->acFileName, (uint8_t*)file_name, str_size);

    file_entry->u32FileSize = file_size;
    file_entry->u32FileAddr = header->u32NextWriteAddr;
    header->u32NextWriteAddr += file_size;

    // Use memmove to accommodate optimizations where the file data is temporarily stored
    // in buffer32
    memmove(((uint8_t*)buffer32) + (file_entry->u32FileAddr), (uint8_t*)file_data, file_size);

    return M2M_SUCCESS;
}


static const char* bin2hex(const void* data, size_t data_size)
{
    static char buf[256];
    static char hex[] = "0123456789abcdef";
    const uint8_t* data8 = data;

    if (data_size * 2 > sizeof(buf) - 1)
    {
        return "[buf too small]";
    }

    for (size_t i = 0; i < data_size; i++)
    {
        buf[i * 2 + 0] = hex[(*data8) >> 4];
        buf[i * 2 + 1] = hex[(*data8) & 0xF];
        data8++;
    }
    buf[data_size * 2] = 0;

    return buf;
}



int8_t transfer_certificates(uint8_t* deviceCert,size_t device_cert_size,uint8_t* signerCert,size_t signer_cert_size)
{
	int qlDebug = 0;
	int8_t status = M2M_SUCCESS;
	int atca_status = ATCACERT_E_SUCCESS;
	
	uint8_t *file_list = NULL;
	char *device_cert_filename = NULL;
	char *signer_cert_filename = NULL;
	uint32_t sector_buffer[MAX_TLS_CERT_LENGTH];
	
	uint8_t cert_sn[CERT_SN_MAX_LEN];
	
	APP_DebugPrintf("Provision WINC TLS Signer & Device Certs (from QL).\r\n\r\n");
	
	do
	{
		// Clear cert chain buffer
		memset(sector_buffer, 0xFF, sizeof(sector_buffer));

		// Use the end of the sector buffer to temporarily hold the data to save RAM
		file_list   = ((uint8_t*)sector_buffer) + (sizeof(sector_buffer) - TLS_FILE_NAME_MAX * 2);


		// Init the file list
		memset(file_list, 0, TLS_FILE_NAME_MAX * 2);
		device_cert_filename = (char*)&file_list[0];
		signer_cert_filename = (char*)&file_list[TLS_FILE_NAME_MAX];

		
		////////////////
		
		// Get the signer certificate SN for the filename
		memcpy(cert_sn,signerCert+15,8);
		
		// Build the signer certificate filename
		strcpy(signer_cert_filename, "CERT_");
		strcat(signer_cert_filename, bin2hex(cert_sn, 8));
		if (qlDebug>0)
		APP_DebugPrintf("DBSG : QL Signer filename '%s'\r\n",signer_cert_filename);
		
		// Add the DER signer certificate the TLS certs buffer
		status = winc_certs_append_file_buf(sector_buffer, sizeof(sector_buffer),
		signer_cert_filename, signer_cert_size, signerCert);
		
		if (status != M2M_SUCCESS)
		{
			// Break the do/while loop
			break;
		}

		
		////////////////////
		
		
		// Get the device certificate SN for the filename
		memcpy(cert_sn,deviceCert+15,8);
		
		
		// Build the device certificate filename
		strcpy(device_cert_filename, "CERT_");
		strcat(device_cert_filename, bin2hex(cert_sn, 8));
		
		if (qlDebug>0)
		APP_DebugPrintf("DBSG : QL Device filename '%s'\r\n",device_cert_filename);
		
		// Add the DER device certificate the TLS certs buffer
		status = winc_certs_append_file_buf(sector_buffer, sizeof(sector_buffer),
		device_cert_filename, device_cert_size,deviceCert);
		
		if (status != M2M_SUCCESS)
		{
			// Break the do/while loop
			break;
		}

		
		
		// Add the cert chain list file to the TLS certs buffer
		status = winc_certs_append_file_buf(sector_buffer, sizeof(sector_buffer),
		TLS_SRV_ECDSA_CHAIN_FILE, TLS_FILE_NAME_MAX * 2, file_list);
		if (status != M2M_SUCCESS)
		{
			// Break the do/while loop
			break;
		}

		file_list = NULL;
		signer_cert_filename = NULL;
		device_cert_filename = NULL;

		// Update the TLS cert chain on the WINC.
		status = m2m_ssl_send_certs_to_winc((uint8_t*)sector_buffer,
		(uint32_t)winc_certs_get_total_files_size((tstrTlsSrvSecHdr*)sector_buffer));
		
		if (status != M2M_SUCCESS)
		{
			// Break the do/while loop
			break;
		}
	}
	while (false);

	if (atca_status)
	{
		M2M_ERR("eccSendCertsToWINC() failed with ret=%d", atca_status);
		status =  M2M_ERR_FAIL;
	}
	
	return status;
}


