/**********************************************
* Copyright (c) 2021-2022 Crypto Quantique Ltd 
***********************************************/


#include "../Inc/qlLibrary.h"
#include <app.h>

#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#define QUARKLINK_OK                              (0)
#define QUARKLINK_ERR_KEY_SIZE_MISMATCH           (10000)
#define QUARKLINK_ERR_ESK_CSR_REQ_OVERFLOW        (10010)
#define QUARKLINK_ERR_CANT_ALLOC_KEYPAIR          (10020)
#define QUARKLINK_ERR_CANT_ALLOC_GROUP            (10030)
#define QUARKLINK_ERR_ZERO_BYTE_DER               (10040)
#define QUARKLINK_ERR_TEMP_CERT_OVERSIZED         (10050)
#define QUARKLINK_ERR_NO_ENDPOINT_DELIM           (10060)
#define QUARKLINK_ERR_BEGIN_CERT_MISSING_1        (10070)
#define QUARKLINK_ERR_END_CERT_MISSING_1          (10080)
#define QUARKLINK_ERR_BEGIN_CERT_MISSING_2        (10090)
#define QUARKLINK_ERR_END_CERT_MISSING_2          (10100)
#define QUARKLINK_ERR_BEGIN_CERT_MISSING_3        (10110)
#define QUARKLINK_ERR_END_CERT_MISSING_3          (10120)
#define QUARKLINK_ERR_ENDPOINT_LF_MISSING         (10130)
#define QUARKLINK_ERR_ENDPOINT_PORT_MISSING       (10140)
#define QUARKLINK_ERR_ENDPOINT_PORT_NUM_MISSING   (10150)
#define QUARKLINK_ERR_ENDPOINT_PORT_NUM_OVERFLOW  (10160)
#define QUARKLINK_ERR_ENDPOINT_PORT_CORRUPT       (10170)
#define QUARKLINK_ERR_INVALID_PARAM               (10180)
#define QUARKLINK_ERR_NOT_INITTED                 (10190)
#define QUARKLINK_ERR_KEY_EXCHANGE_NOT_ACHIEVED   (10200)
#define QUARKLINK_ERR_BAD_STATUS                  (10210)
#define QUARKLINK_ERR_REQ_OVERFLOW                (10220)

/********************************************************/
/* DSK certificate */
const  char      *pcDskCert          ;
/* The Just In Time Programming/Registration Certificate */
const char       *pcJitpCert         ;
/* Endpoint server name */
const char       *pcEndpointName     ;
/* Port number for endpoint server */
uint16_t          usEndpointPort	 ;

/* Endpoint server certificate. The non-NULL-ness of this variable
 * indicates key exchange succeeded
 * */
const char       *pcEndpointCert     ;








/** Parse successful "simpleenrol" response data in pucRespBody[]
 *
 * @param[in]   pcBody      Pointer to message body text
 * @return 0 on success, non-0 on error
 */
int parseEnrolResponse(char *pcBody, struct ProvisionedDetails *pPdOut)
{
  int                 ret                 = QUARKLINK_OK;

  /* Cast off the unsignedness to make easier to use with str*() fns */
  const size_t        xBeginCertLen       = sizeof("-----BEGIN CERTIFICATE-----\n") - 1;
  const size_t        xEndCertLen         = sizeof("-----END CERTIFICATE-----\n") - 1;
  const size_t        xEndpointDelimLen   = sizeof("\n\nendpoint:") - 1;
  const size_t        xEndpointPortLen    = sizeof("endpointPort:") - 1;

  char                *pc                 = NULL;
  char                *pc2                = NULL;

  ret = strncmp(pcBody, "-----BEGIN CERTIFICATE-----\n", sizeof("-----BEGIN CERTIFICATE-----\n")-1);
  if  (ret != 0)
  {
    ret = QUARKLINK_ERR_BEGIN_CERT_MISSING_1;
    APP_DebugPrintf ("** HTTPS: SE resp body doesn't start with a cert\r\n");
  }

  if (0 == ret)
  {
    pc = strstr(pcBody, "\n\nendpoint:");
	if (NULL == pc)
	{
      ret = QUARKLINK_ERR_NO_ENDPOINT_DELIM;
      APP_DebugPrintf ("** HTTPS: no endpoint delimiter\r\n");
	}
  }

  /* If we find the start of the endpoint, we've found the end of the DSK,
   * and we can NUL-terminate that string.
   */
  if (0 == ret)
  {
	/* Store endpoint position here (actually at first '\n') */
	pc2 = pc;
	/* NUL-terminate the DSK certificate, overwriting second '\n' */
	pc[1] = '\0';
	/* Found the DSK certificate now */
	pcDskCert = pcBody;

	strcpy(pPdOut->dcrt, pcDskCert);

	/* Double-check we've two certs in the DSK */
	pc = pcBody + xBeginCertLen;
	pc = strstr(pc, "-----END CERTIFICATE-----\n");
	if (NULL == pc)
	{
      ret = QUARKLINK_ERR_END_CERT_MISSING_1;
      APP_DebugPrintf ("** HTTPS: can't find end of first cert");
	}
  }

  if (0 == ret)
  {
	pc += xEndCertLen;
	pc = strstr(pc, "-----BEGIN CERTIFICATE-----\n");
	if (NULL == pc)
	{
      ret =QUARKLINK_ERR_BEGIN_CERT_MISSING_2;
      APP_DebugPrintf ("** HTTPS: can't find start of second cert\r\n");
	}
  }

  if (0 == ret)
  {
	pcJitpCert = pc;

	pc += xBeginCertLen;
	pc = strstr(pc, "-----END CERTIFICATE-----\n");
	if (NULL == pc)
	{
      ret =QUARKLINK_ERR_END_CERT_MISSING_2;
      APP_DebugPrintf ("** HTTPS: can't find end of second cert");
	}
  }

  /* Found DCRT data are valid,  parse the endpoint. */

  if (0 == ret)
  {
	/* Skip the prefix to get to the start of the name */
	pc2 += xEndpointDelimLen;

	/* Look for the end of the endpoint name */
	pc = strchr(pc2, '\n');
	if (NULL == pc)
	{
      ret =QUARKLINK_ERR_ENDPOINT_LF_MISSING;
      APP_DebugPrintf ("** HTTPS: can't find LF at end of endpoint");
	}
  }

  if (0 == ret)
  {
	/* Found endpoint name  */
	pcEndpointName = pc2;

	/* NUL-terminate the endpoint name, overwriting the LF */
	*pc = '\0';
	/* The next character should start "endpointPort:" */
	++pc;    /* Skip line end */

	strcpy(pPdOut->iotHubURL, pcEndpointName);

	int rc = strncmp(pc, "endpointPort:", xEndpointPortLen);
	if (rc != 0)
    {
      ret =QUARKLINK_ERR_ENDPOINT_PORT_MISSING;
      APP_DebugPrintf ("** HTTPS: endpoint port header missing\r\n");
    }
  }

  unsigned int uPortNo = 0UL;
  char c;

  if (0 == ret)
  {
	/* Skip past to the start of the port number itself */
	pc += xEndpointPortLen;

    do
	{
      c = *(pc++);

      if (isdigit(c))
      {
        if(uPortNo > 65536)
        {
          ret =QUARKLINK_ERR_ENDPOINT_PORT_NUM_OVERFLOW;
          APP_DebugPrintf ("** HTTPS: endpoint port number corrupt\r\n");
        }
        uPortNo = (uPortNo * 10) + (unsigned int)(c - '0');

      }
      else if ('\n' != c)
      {
        ret = QUARKLINK_ERR_ENDPOINT_PORT_CORRUPT;
        APP_DebugPrintf ("** HTTPS: corruption after endpoint port number: found %c (0x%x)\r\n", (isprint(c) ? c : ' '), c);
      }

	} while (0 == ret && '\n' != c);
  }

  if (0 == ret)
  {
    if(0u == uPortNo)
    {
      ret =QUARKLINK_ERR_ENDPOINT_PORT_NUM_MISSING;
      APP_DebugPrintf ("** HTTPS: endpoint port number missing or zero\r\n");
    }
  }

  if (0 == ret)
  {
	if(65536 < uPortNo)
	{
      ret =QUARKLINK_ERR_ENDPOINT_PORT_NUM_OVERFLOW;
      APP_DebugPrintf ("** HTTPS: endpoint port number overflow\r\n");
	}
  }

  /* parse completed */
  if (0 == ret)
  {
    usEndpointPort = (uint16_t)uPortNo;

	pPdOut->iotHubPort = usEndpointPort;

	pc = strstr(pc, "-----BEGIN CERTIFICATE-----\n");
	if(NULL == pc)
	{
      ret =QUARKLINK_ERR_BEGIN_CERT_MISSING_3;
      APP_DebugPrintf ("** HTTPS: can't find start of third cert\r\n");
	}
	/* Note the start of the begin certificate line for later */
	pc2 = pc;
  }

  if (0 == ret)
  {
	pc += xBeginCertLen;
	/* note the last Cert may not have CR on the end */
	pc = strstr(pc, "-----END CERTIFICATE-----");
	if(NULL == pc)
	{
      ret =QUARKLINK_ERR_END_CERT_MISSING_3;
      APP_DebugPrintf ("** HTTPS: can't find end of third cert\r\n");
	}
  }

  if (0 == ret)
  {
    /* Server certificate's validated, so store its location. This also
	 * flags that all the key exchange has succeeded.
	 */
    pcEndpointCert = pc2;
	pc[xEndCertLen] = '\0';
	strcpy(pPdOut->iotHubCert, pcEndpointCert);
  }

  return ret;
}



uint32_t parse(char *inStr,char *startStr,char *endStr,int radix)
{
 char *pch1 = NULL;
 char *pch2 = NULL;
 char temp[3000];
 
 uint32_t ret;
 
 strcpy(temp,inStr);
 
 pch1 = strstr(temp,startStr);
    if (pch1)
        {
        pch1 += strlen(startStr);
        if (endStr[0]!='\0'){
            pch2 = strstr(pch1,endStr);
            *pch2 = '\0';
            }
        
        switch(radix)
            {
            case 0 : // return pointer
                ret = (uint32_t)&pch1;
                return ret;
                break;
            case 1 : // return intValue extract
                return atoi(pch1);
                break;
            }
        }
 return 0;
}
            